/****************************************************************************

 KHotKeys
 
 Copyright (C) 1999-2001 Lubos Lunak <l.lunak@kde.org>

 Distributed under the terms of the GNU General Public License version 2.
 
 $Id: conditions.cpp,v 1.3 2001/07/15 14:08:29 seli Exp $

****************************************************************************/

#define __CONDITIONS_CPP

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "conditions.h"

#include <kconfig.h>
#include <kdebug.h>
#include <klocale.h>
#include <assert.h>

#include <X11/Xlib.h>
#include <fixx11h.h>

#include <action_data.h>

namespace KHotKeys
{

// Condition

Condition* Condition::create_cfg_read( KConfig& cfg_P, const Condition* parent_P )
    {
    QString type = cfg_P.readEntry( "Type" );
    if( type == "ACTIVE_WINDOW" )
        return new Active_window_condition( cfg_P, parent_P );
    if( type == "EXISTING_WINDOW" )
        return new Existing_window_condition( cfg_P, parent_P );
    if( type == "NOT" )
        return new Not_condition( cfg_P, parent_P );
    kdWarning( 1213 ) << "Unknown Condition type read from cfg file\n";
    return NULL;
    }
    
void Condition::cfg_write( KConfig& cfg_P ) const
    {
    cfg_P.writeEntry( "Type", "ERROR" );
    }
    
void Condition::updated() const
    {
    if( !khotkeys_active())
        return;
    assert( parent != NULL );
    parent->updated();
    }

// Condition_list_base

Condition_list_base::Condition_list_base( KConfig& cfg_P, const Condition* parent_P )
    : Condition( parent_P )
    {
    setAutoDelete( true );
    QString save_cfg_group = cfg_P.group();
    int cnt = cfg_P.readNumEntry( "ConditionsCount", 0 );
    for( int i = 0;
         i < cnt;
         ++i )
        {
        cfg_P.setGroup( save_cfg_group + QString::number( i ));
        Condition* condition = Condition::create_cfg_read( cfg_P, this );
        if( condition )
            append( condition );
        }
    cfg_P.setGroup( save_cfg_group );
    }
    
void Condition_list_base::cfg_write( KConfig& cfg_P ) const
    {
    QString save_cfg_group = cfg_P.group();
    int i = 0;
    for( Iterator it( *this );
         it;
         ++it, ++i )
        {
        cfg_P.setGroup( save_cfg_group + QString::number( i ));
        it.current()->cfg_write( cfg_P );
        }
    cfg_P.setGroup( save_cfg_group );
    cfg_P.writeEntry( "ConditionsCount", i );     
    }

// Condition_list

Condition_list::Condition_list( KConfig& cfg_P, Action_data_base* data_P )
    : Condition_list_base( cfg_P, NULL ), data( data_P )
    {
    _comment = cfg_P.readEntry( "Comment" );
    }

void Condition_list::cfg_write( KConfig& cfg_P ) const
    {
    base::cfg_write( cfg_P );
    cfg_P.writeEntry( "Comment", comment());     
    }

Condition_list* Condition_list::copy( Action_data_base* data_P ) const
    {
    Condition_list* ret = new Condition_list( comment(), data_P );
    for( Iterator it( *this );
         it;
         ++it )
        ret->append( it.current()->copy( this ));
    return ret;
    }
    
    
bool Condition_list::match() const
    {
    if( count() == 0 ) // no conditions to match => ok
        return true;
    for( Iterator it( *this );
         it;
         ++it )
        if( it.current()->match()) // CHECKME OR ( nebo AND ) ?
            return true;
    return false;
    }
    
void Condition_list::updated() const
    {
    if( !khotkeys_active())
        return;
    data->update_triggers();
    }

// CHECKME tohle je drobet hack, jeste to zvazit
void Condition_list::set_data( Action_data_base* data_P )
    {
    assert( data == NULL || data == data_P );
    data = data_P;
    }

const QString Condition_list::description() const
    {
    assert( false );
    }

Condition_list* Condition_list::copy( const Condition* ) const
    {
    assert( false );
    }

// Active_window_condition

Active_window_condition::Active_window_condition( KConfig& cfg_P, const Condition* parent_P )
    : Condition( cfg_P, parent_P )
    {
    QString save_cfg_group = cfg_P.group();
    cfg_P.setGroup( save_cfg_group + "Window" );
    _window = new Windowdef_list( cfg_P );
    cfg_P.setGroup( save_cfg_group );
    init();
    set_match();
    }

void Active_window_condition::init()
    {
    connect( windows_handler, SIGNAL( active_window_changed( WId )),
        this, SLOT( active_window_changed( WId )));
    }
            
bool Active_window_condition::match() const
    {
    return is_match;
    }
    
void Active_window_condition::set_match()
    {
    is_match = window()->match( Window_data( windows_handler->active_window()));
    kdDebug( 1213 ) << "Active_window_condition::set_match :" << is_match << endl;
    updated();
    }

void Active_window_condition::cfg_write( KConfig& cfg_P ) const
    {
    base::cfg_write( cfg_P );
    QString save_cfg_group = cfg_P.group();
    cfg_P.setGroup( save_cfg_group + "Window" );
    window()->cfg_write( cfg_P );
    cfg_P.setGroup( save_cfg_group );
    cfg_P.writeEntry( "Type", "ACTIVE_WINDOW" ); // overwrites value set in base::cfg_write()
    }

#ifndef RETURN_TYPE_RELAXATION_BROKEN
Active_window_condition* Active_window_condition::copy( const Condition* parent_P ) const
#else
Condition* Active_window_condition::copy( const Condition* parent_P ) const
#endif
    {
    return new Active_window_condition( window()->copy(), parent_P );
    }
    
const QString Active_window_condition::description() const
    {
    return i18n( "Active window : " ) + window()->comment();
    }

void Active_window_condition::active_window_changed( WId )
    {
    set_match();
    }
    
Active_window_condition::~Active_window_condition()
    {
    disconnect( windows_handler, NULL, this, NULL );
    delete _window;
    }
    
// Existing_window_condition

Existing_window_condition::Existing_window_condition( KConfig& cfg_P, const Condition* parent_P )
    : Condition( cfg_P, parent_P )
    {
    QString save_cfg_group = cfg_P.group();
    cfg_P.setGroup( save_cfg_group + "Window" );
    _window = new Windowdef_list( cfg_P );
    cfg_P.setGroup( save_cfg_group );
    init();
    set_match();
    }
    
void Existing_window_condition::init()
    {
    connect( windows_handler, SIGNAL( window_added( WId )), this, SLOT( window_added( WId )));
    connect( windows_handler, SIGNAL( window_removed( WId )), this, SLOT( window_removed( WId )));
    }
    
bool Existing_window_condition::match() const
    {
    return is_match;
    }
    
void Existing_window_condition::set_match( WId w_P )
    {
    if( w_P != None && !is_match )
        is_match = window()->match( Window_data( w_P ));
    else
        is_match = windows_handler->find_window( window()) != None;
    kdDebug( 1213 ) << "Existing_window_condition::set_match :" << is_match << endl;
    updated();
    }

void Existing_window_condition::cfg_write( KConfig& cfg_P ) const
    {
    base::cfg_write( cfg_P );
    QString save_cfg_group = cfg_P.group();
    cfg_P.setGroup( save_cfg_group + "Window" );
    window()->cfg_write( cfg_P );
    cfg_P.setGroup( save_cfg_group );
    cfg_P.writeEntry( "Type", "EXISTING_WINDOW" ); // overwrites value set in base::cfg_write()
    }

#ifndef RETURN_TYPE_RELAXATION_BROKEN
Existing_window_condition* Existing_window_condition::copy( const Condition* parent_P ) const
#else
Condition* Existing_window_condition::copy( const Condition* parent_P ) const
#endif
    {
    return new Existing_window_condition( window()->copy(), parent_P );
    }
    
const QString Existing_window_condition::description() const
    {
    return i18n( "Existing window : " ) + window()->comment();
    }

void Existing_window_condition::window_added( WId w_P )
    {
    set_match( w_P );
    }
    
void Existing_window_condition::window_removed( WId )
    {
    set_match();
    }

Existing_window_condition::~Existing_window_condition()
    {
    disconnect( windows_handler, NULL, this, NULL );
    delete _window;
    }
    
// Not_condition

Not_condition::Not_condition( KConfig& cfg_P, const Condition* parent_P )
    : Condition_list_base( cfg_P, parent_P )
    {
    // CHECKME kontrola poctu ?
    }

bool Not_condition::match() const
    {
    return condition() ? !condition()->match() : false;
    }
    
void Not_condition::cfg_write( KConfig& cfg_P ) const
    {
    base::cfg_write( cfg_P );
    cfg_P.writeEntry( "Type", "NOT" ); // overwrites value set in base::cfg_write()
    }

Not_condition* Not_condition::copy( const Condition* parent_P ) const
    {
    return new Not_condition( condition() ? condition()->copy( this ) : NULL, parent_P );
    }
    
const QString Not_condition::description() const
    {
    return i18n( "Not_condition", "Not" );
    }

} // namespace KHotKeys

#include "conditions.moc"
