/****************************************************************************

 KHotKeys
 
 Copyright (C) 1999-2001 Lubos Lunak <l.lunak@kde.org>

 Distributed under the terms of the GNU General Public License version 2.
 
 $Id: triggers.h,v 1.17 2001/11/20 14:00:11 seli Exp $

****************************************************************************/

#ifndef __TRIGGERS_H
#define __TRIGGERS_H

#include <qlist.h>
#include <qtimer.h>
#include <qmap.h>

#include <khotkeysglobal.h>

#include "input.h"

class KConfig;

namespace KHotKeys
{

class Windowdef_list;
class Action_data;

class Trigger
    {
    public:
        Trigger( Action_data* data_P );
        Trigger( KConfig& cfg_P, Action_data* data_P );
        virtual ~Trigger();
        virtual void cfg_write( KConfig& cfg_P ) const = 0;
        virtual Trigger* copy( Action_data* data_P ) const = 0;
        virtual const QString description() const = 0;
        static Trigger* create_cfg_read( KConfig& cfg_P, Action_data* data_P );
        virtual void activate( bool activate_P ) = 0;
    protected:
        Action_data* const data;
    KHOTKEYS_DISABLE_COPY( Trigger );
    };

class Trigger_list
    : public QList< Trigger >
    {
    public:
        Trigger_list( const QString& comment_P ); // CHECKME nebo i data ?
        Trigger_list( KConfig& cfg_P, Action_data* data_P );
        void activate( bool activate_P );
        void cfg_write( KConfig& cfg_P ) const;
        typedef QListIterator< Trigger > Iterator;
        const QString& comment() const;
        Trigger_list* copy( Action_data* data_P ) const;
    private:
        QString _comment;
    KHOTKEYS_DISABLE_COPY( Trigger_list );
    };
    
class Single_shortcut_trigger
    : public Trigger, public Kbd_receiver
    {
    typedef Trigger base;
    public:
        Single_shortcut_trigger( Action_data* data_P, unsigned int keycode_P );
        Single_shortcut_trigger( KConfig& cfg_P, Action_data* data_P );
        virtual ~Single_shortcut_trigger();
        virtual void cfg_write( KConfig& cfg_P ) const;
        virtual Single_shortcut_trigger* copy( Action_data* data_P ) const;
        virtual const QString description() const;
        unsigned int keycode() const;
        virtual bool handle_key( unsigned int keycode_P ); // CHECKME nebo slot ?
        virtual void activate( bool activate_P );
    private:
        unsigned int _keycode;
    };

#ifdef DOUBLE_SHORTCUT
class Double_shortcut_trigger
    : public Shortcut_trigger
    {
// CHECKME otherwise moc generates code for this class ( even when #if 0 )
//    Q_OBJECT
    typedef Shortcut_trigger base;
    public:
        Double_shortcut_trigger( unsigned int keycode1_P, unsigned int keycode2_P,
            unsigned int after_first_timeout_P, Action_data* data_P );
        Double_shortcut_trigger( KConfig& cfg_P, Action_data* data_P );
        virtual void cfg_write( KConfig& cfg_P ) const;
        virtual Double_shortcut_trigger* copy( Action_data* data_P ) const;
        unsigned int keycode1() const;
        unsigned int keycode2() const;
        unsigned int after_first_timeout() const;
    protected:
        virtual void activated( unsigned int keycode_P, bool& handled_P );
        bool after_first;
        QTimer after_first_timer;
    private:
        unsigned int _keycode1, _keycode2;
        unsigned int _after_first_timeout;
    };
#endif

class Gesture_trigger
    : public QObject, public Trigger
    {
    Q_OBJECT
    typedef Trigger base;
    public:
        Gesture_trigger( Action_data* data_P, const QString &gesturecode_P );
        Gesture_trigger( KConfig& cfg_P, Action_data* data_P );
        virtual ~Gesture_trigger();
        virtual void cfg_write( KConfig& cfg_P ) const;
        virtual Trigger* copy( Action_data* data_P ) const;
        virtual const QString description() const;
        inline const QString &gesturecode() const { return _gesturecode; }
        inline void setGesturecode(const QString &code) { _gesturecode = code; }
        virtual void activate( bool activate_P );
        
    protected slots:
        void handle_gesture(const QString &gesture);
        
    private:
        QString _gesturecode;
        bool _active;
    };

class Window_trigger
    : public QObject, public Trigger        
    {
    Q_OBJECT
    typedef Trigger base;
    public:
        enum window_action_t
            {
            WINDOW_APPEARS         = ( 1 << 0 ),
            WINDOW_DISAPPEARS      = ( 1 << 1 ),
            WINDOW_ACTIVATES       = ( 1 << 2 ),
            WINDOW_DEACTIVATES     = ( 1 << 3 )
            };
        Window_trigger( Action_data* data_P, Windowdef_list* windows_P, int window_actions_P );
        Window_trigger( KConfig& cfg_P, Action_data* data_P );
        virtual ~Window_trigger();
        virtual void cfg_write( KConfig& cfg_P ) const;
#ifndef RETURN_TYPE_RELAXATION_BROKEN    // stupid gcc, it doesn't even warn it can't do this
        virtual Window_trigger* copy( Action_data* data_P ) const;
#else
        virtual Trigger* copy( Action_data* data_P ) const;
#endif
        virtual const QString description() const;
        const Windowdef_list* windows() const;
        bool triggers_on( window_action_t w_action_P ) const;
        virtual void activate( bool activate_P );
    protected: // CHECKME neco private ?
        Windowdef_list* _windows;
        int window_actions;
        void init();
        typedef QMap< WId, bool > Windows_map;
        Windows_map existing_windows;
        WId last_active_window;
    protected slots:
        void window_added( WId window_P );
        void window_removed( WId window_P );
        void active_window_changed( WId window_P );
//        void window_changed( WId window_P, unsigned int dirty_P );
    protected:
        bool active;
    };

//***************************************************************************
// Inline
//***************************************************************************

// Trigger

inline
Trigger::Trigger( Action_data* data_P )
    : data( data_P )
    {
    }

inline
Trigger::Trigger( KConfig&, Action_data* data_P )
    : data( data_P )
    {
    }
        
inline
Trigger::~Trigger()
    {
    }

// Trigger_list
    
inline
Trigger_list::Trigger_list( const QString& comment_P )
    : QList< Trigger >(), _comment( comment_P )
    {
    setAutoDelete( true );
    }

inline
const QString& Trigger_list::comment() const
    {
    return _comment;
    }
    
// Single_shortcut_trigger

inline
unsigned int Single_shortcut_trigger::keycode() const
    {
    return _keycode;
    }

#ifdef DOUBLE_SHORTCUT
// Double_shortcut_trigger

inline
unsigned int Double_shortcut_trigger::keycode1() const
    {
    return _keycode1;
    }

inline
unsigned int Double_shortcut_trigger::keycode2() const
    {
    return _keycode2;
    }

inline
unsigned int Double_shortcut_trigger::after_first_timeout() const
    {
    return _after_first_timeout;
    }
#endif

// Window_trigger

inline
Window_trigger::Window_trigger( Action_data* data_P, Windowdef_list* windows_P,
     int window_actions_P )
    : Trigger( data_P ), _windows( windows_P ), window_actions( window_actions_P ),
      last_active_window( None ), active( false )
    {
    init();
    }
    
inline
const Windowdef_list* Window_trigger::windows() const
    {
    return _windows;
    }
    
inline
bool Window_trigger::triggers_on( window_action_t w_action_P ) const
    {
    return window_actions & w_action_P;
    }

} // namespace KHotKeys
    
#endif
